@@ -178,10 +178,10 @@ class SklearnIntentClassifier(Component):
         else:
             X = message.get("text_features").reshape(1, -1)
             intent_ids, probabilities = self.predict(X)
-            intents = self.transform_labels_num2str(intent_ids)
+            intents = self.transform_labels_num2str(np.ravel(intent_ids))
             # `predict` returns a matrix as it is supposed
             # to work for multiple examples as well, hence we need to flatten
-            intents, probabilities = intents.flatten(), probabilities.flatten()
+            probabilities = probabilities.flatten()
 
             if intents.size > 0 and probabilities.size > 0:
                 ranking = list(zip(list(intents),
